# errors.py
"""
Custom error handling for the XQueue submission interface.
"""


class XQueueSubmissionError(Exception):
    """Base class for all XQueue submission errors."""


class JSONParsingError(XQueueSubmissionError):
    """Raised when JSON parsing fails."""

    MESSAGE = "Error parsing {name}: {error}"

    def __init__(self, name, error):
        super().__init__(self.MESSAGE.format(name=name, error=error))


class MissingKeyError(XQueueSubmissionError):
    """Raised when a required key is missing."""

    MESSAGE = "Missing key: {key}"

    def __init__(self, key):
        super().__init__(self.MESSAGE.format(key=key))


class ValidationError(XQueueSubmissionError):
    """Raised when a validation check fails."""

    MESSAGE = "Validation error: {error}"

    def __init__(self, error):
        super().__init__(self.MESSAGE.format(error=error))


class TypeErrorSubmission(XQueueSubmissionError):
    """Raised when an invalid type is encountered."""

    MESSAGE = "Type error: {error}"

    def __init__(self, error):
        super().__init__(self.MESSAGE.format(error=error))


class RuntimeErrorSubmission(XQueueSubmissionError):
    """Raised for runtime errors."""

    MESSAGE = "Runtime error: {error}"

    def __init__(self, error):
        super().__init__(self.MESSAGE.format(error=error))


class GetSubmissionParamsError(XQueueSubmissionError):
    """Raised when there is an issue getting submission parameters."""

    MESSAGE = "Submission parameters error: {error}"

    def __init__(self, error="Block instance is not defined!"):
        super().__init__(self.MESSAGE.format(error=error))



"""A module proxy for delayed importing of modules.

From http://barnesc.blogspot.com/2006/06/automatic-python-imports-with-autoimp.html,
in the public domain.

"""

import sys


class LazyModule:
    """A lazy module proxy."""

    def __init__(self, modname):
        self.__dict__["__name__"] = modname
        self._set_mod(None)

    def _set_mod(self, mod):
        if mod is not None:
            self.__dict__ = mod.__dict__
        self.__dict__["_lazymod_mod"] = mod

    def _load_mod(self):
        __import__(self.__name__)
        self._set_mod(sys.modules[self.__name__])

    def __getattr__(self, name):
        if self.__dict__["_lazymod_mod"] is None:
            self._load_mod()

        mod = self.__dict__["_lazymod_mod"]

        if hasattr(mod, name):
            return getattr(mod, name)

        try:
            subname = f"{self.__name__}.{name}"
            __import__(subname)
            submod = getattr(mod, name)  # pylint: disable=unused-variable
        except ImportError as exc:
            raise AttributeError(f"'module' object has no attribute {name!r}") from exc
        self.__dict__[name] = LazyModule(subname)
        return self.__dict__[name]



"""Poll block is ungraded xmodule used by students to
to do set of polls.

On the client side we show:
If student does not yet anwered - Question with set of choices.
If student have answered - Question with statistics for each answers.
"""

import json
import logging
from collections import OrderedDict
from copy import deepcopy

import markupsafe
from django.utils.translation import gettext_noop as _
from lxml import etree
from opaque_keys.edx.keys import UsageKey
from web_fragments.fragment import Fragment
from xblock.core import XBlock
from xblock.fields import Boolean, Dict, List, Scope, String
from xblock.utils.resources import ResourceLoader

from xblocks_contrib.common.xml_utils import LegacyXmlMixin

Text = markupsafe.escape
resource_loader = ResourceLoader(__name__)
log = logging.getLogger(__name__)


def HTML(html):  # pylint: disable=invalid-name
    """
    Mark a string as already HTML, so that it won't be escaped before output.

    Use this function when formatting HTML into other strings.  It must be
    used in conjunction with ``Text()``, and both ``HTML()`` and ``Text()``
    must be closed before any calls to ``format()``::

        <%page expression_filter="h"/>
        <%!
        from django.utils.translation import gettext as _

        from openedx.core.djangolib.markup import HTML, Text
        %>
        ${Text(_("Write & send {start}email{end}")).format(
            start=HTML("<a href='mailto:{}'>").format(user.email),
            end=HTML("</a>"),
        )}

    """
    return markupsafe.Markup(html)


def stringify_children(node):
    """
    Return all contents of an xml tree, without the outside tags.
    e.g. if node is parse of

        "<html a="b" foo="bar">Hi <div>there <span>Bruce</span><b>!</b></div><html>"

    should return

        "Hi <div>there <span>Bruce</span><b>!</b></div>"

    fixed from
    http://stackoverflow.com/questions/4624062/get-all-text-inside-a-tag-in-lxml
    """
    # Useful things to know:

    # node.tostring() -- generates xml for the node, including start
    #                 and end tags.  We'll use this for the children.
    # node.text -- the text after the end of a start tag to the start
    #                 of the first child
    # node.tail -- the text after the end this tag to the start of the
    #                 next element.
    parts = [node.text]
    for c in node.getchildren():
        parts.append(etree.tostring(c, with_tail=True, encoding="unicode"))

    # filter removes possible Nones in texts and tails
    return "".join([part for part in parts if part])


@XBlock.needs("i18n")
class PollBlock(LegacyXmlMixin, XBlock):
    """Poll Block"""

    # Indicates that this XBlock has been extracted from edx-platform.
    is_extracted = True

    # Name of poll to use in links to this poll
    display_name = String(
        help=_("The display name for this component."),
        scope=Scope.settings,
        default="poll_question",
    )

    voted = Boolean(help=_("Whether this student has voted on the poll"), scope=Scope.user_state, default=False)
    poll_answer = String(help=_("Student answer"), scope=Scope.user_state, default="")
    poll_answers = Dict(help=_("Poll answers from all students"), scope=Scope.user_state_summary)

    # List of answers, in the form {'id': 'some id', 'text': 'the answer text'}
    answers = List(help=_("Poll answers from xml"), scope=Scope.content, default=[])

    question = String(help=_("Poll question"), scope=Scope.content, default="")

    js_module_name = "poll"

    _tag_name = "poll_question"
    _child_tag_name = "answer"

    @property
    def xblock_kvs(self):
        """
        Retrieves the internal KeyValueStore for this XModule.

        Should only be used by the persistence layer. Use with caution.
        """
        # if caller wants kvs, caller's assuming it's up to date; so, decache it
        self.save()
        return self._field_data._kvs  # pylint: disable=protected-access

    @property
    def url_name(self):
        return self.location.block_id

    @property
    def course_id(self):
        return self.location.course_key

    @property
    def category(self):
        return self.scope_ids.block_type

    @property
    def location(self):
        return self.scope_ids.usage_id

    @location.setter
    def location(self, value):
        assert isinstance(value, UsageKey)
        self.scope_ids = self.scope_ids._replace(
            def_id=value,  # Note: assigning a UsageKey as def_id is OK in old mongo / import system but wrong in split
            usage_id=value,
        )

    def handle_ajax(self, dispatch, data):  # legacy support for tests
        """
        Legacy method to mimic old ajax handler behavior for backward compatibility.
        """
        if dispatch == "get_state":
            return json.dumps(self.handle_get_state(data))
        else:
            return json.dumps(self.submit_answer(dispatch))

    def student_view(self, _context):
        """
        Renders the student view.
        """

        frag = Fragment()
        frag.add_content(
            resource_loader.render_django_template(
                "templates/poll.html",
                {
                    "element_id": self.scope_ids.usage_id.html_id(),
                    "element_class": self.scope_ids.usage_id.block_type,
                    "configuration_json": self.dump_poll(),
                },
                i18n_service=self.runtime.service(self, "i18n"),
            )
        )
        frag.add_css(resource_loader.load_unicode("static/css/poll.css"))
        frag.add_javascript(resource_loader.load_unicode("static/js/src/poll.js"))
        frag.initialize_js("PollBlock")
        return frag

    def dump_poll(self):
        """Dump poll information.

        Returns:
            string - Serialize json.
        """
        # FIXME: hack for resolving caching `default={}` during definition
        # poll_answers field

        if self.poll_answers is None:
            self.poll_answers = {}

        answers_to_json = OrderedDict()

        # # # FIXME: fix this, when xblock support mutable types.
        # # # Now we use this hack.
        temp_poll_answers = self.poll_answers

        # # # Fill self.poll_answers, prepare data for template context.
        for answer in self.answers:
            # Set default count for answer = 0.
            if answer["id"] not in temp_poll_answers:
                temp_poll_answers[answer["id"]] = 0
            answers_to_json[answer["id"]] = answer["text"]
        self.poll_answers = temp_poll_answers

        return json.dumps(
            {
                "answers": answers_to_json,
                "question": self.question,
                "poll_answer": self.poll_answer,
                "poll_answers": self.poll_answers,
                "total": sum(self.poll_answers.values()) if self.voted else 0,
                "reset": str(self.xml_attributes.get("reset", "true")).lower(),
            }
        )

    @XBlock.json_handler
    def handle_get_state(self, data, suffix=""):  # pylint: disable=unused-argument
        return {
            "poll_answer": self.poll_answer,
            "poll_answers": self.poll_answers,
            "total": sum(self.poll_answers.values()),
        }

    def submit_answer(self, answer):
        """
        Submits a poll answer.
        """
        if not answer:
            return {"error": "No answer provided!"}

        if answer in self.poll_answers and not self.voted:
            # FIXME: fix this, when xblock will support mutable types.
            # Now we use this hack.
            temp_poll_answers = self.poll_answers
            temp_poll_answers[answer] += 1
            self.poll_answers = temp_poll_answers

            self.voted = True
            self.poll_answer = answer
            return {
                "poll_answers": self.poll_answers,
                "total": sum(self.poll_answers.values()),
                "callback": {"objectName": "Conditional"},
            }
        return {"error": "Unknown Command!"}

    @XBlock.json_handler
    def handle_submit_state(self, data, suffix=""):  # pylint: disable=unused-argument
        """
        handler to submit poll answer.
        """
        answer = data.get("answer")  # Extract the answer from the data payload
        return self.submit_answer(answer)

    @XBlock.json_handler
    def handle_reset_state(self):
        """
        handler to Reset poll answer.
        """

        self.voted = False

        # FIXME: fix this, when xblock will support mutable types.
        # Now we use this hack.
        temp_poll_answers = self.poll_answers
        temp_poll_answers[self.poll_answer] -= 1
        self.poll_answers = temp_poll_answers
        self.poll_answer = ""
        return {"status": "success"}

    @staticmethod
    def workbench_scenarios():
        """Create canned scenario for display in the workbench."""
        return [
            (
                "PollBlock",
                """<_poll_question_extracted/>
                """,
            ),
            (
                "Multiple PollBlock",
                """<vertical_demo>
                <_poll_question_extracted/>
                <_poll_question_extracted/>
                <_poll_question_extracted/>
                </vertical_demo>
                """,
            ),
        ]

    def get_explicitly_set_fields_by_scope(self, scope=Scope.content):
        """
        Get a dictionary of the fields for the given scope which are set explicitly on this xblock. (Including
        any set to None.)
        """
        result = {}
        for field in self.fields.values():
            if field.scope == scope and field.is_set_on(self):
                try:
                    result[field.name] = field.read_json(self)
                except TypeError as exception:
                    exception_message = "{message}, Block-location:{location}, Field-name:{field_name}".format(
                        message=str(exception), location=str(self.location), field_name=field.name
                    )
                    raise TypeError(exception_message)  # pylint: disable=raise-missing-from
        return result

    @classmethod
    def definition_from_xml(cls, xml_object, system):
        """
        Pull out the data into a dictionary.

        Args:
            xml_object: XML from file.
            system: `system` object.

        Returns:
            tuple: A tuple ``(definition, children)``.

            definition (dict):
                A dictionary containing:

                - ``answers`` (list): List of answers.
                - ``question`` (str): Question string.
        """
        # Check for presense of required tags in xml.
        if len(xml_object.xpath(cls._child_tag_name)) == 0:
            raise ValueError(
                "Poll_question definition must include \
                at least one 'answer' tag"
            )

        xml_object_copy = deepcopy(xml_object)
        answers = []
        for element_answer in xml_object_copy.findall(cls._child_tag_name):
            answer_id = element_answer.get("id", None)
            if answer_id:
                answers.append({"id": answer_id, "text": stringify_children(element_answer)})
            xml_object_copy.remove(element_answer)

        definition = {"answers": answers, "question": stringify_children(xml_object_copy)}
        children = []
        return (definition, children)

    def definition_to_xml(self, resource_fs=None):
        """Return an xml element representing to this definition."""

        poll_str = HTML("<{tag_name}>{text}</{tag_name}>").format(tag_name=self._tag_name, text=self.question)
        xml_object = etree.fromstring(poll_str)
        xml_object.set("display_name", self.display_name)

        def add_child(xml_obj, answer):  # pylint: disable=unused-argument
            # Escape answer text before adding to xml tree.
            answer_text = str(answer["text"])
            child_str = Text("{tag_begin}{text}{tag_end}").format(
                tag_begin=HTML('<{tag_name} id="{id}">').format(tag_name=self._child_tag_name, id=answer["id"]),
                text=answer_text,
                tag_end=HTML("</{tag_name}>").format(tag_name=self._child_tag_name),
            )
            child_node = etree.fromstring(child_str)
            xml_object.append(child_node)

        for answer in self.answers:
            add_child(xml_object, answer)

        return xml_object



# coding=utf-8
"""
Tests capa util
"""

import unittest

import codejail.safe_exec
import ddt
from django.test.utils import TestContextDecorator
from lxml import etree

from xblocks_contrib.problem.capa.tests.helpers import mock_capa_system
from xblocks_contrib.problem.capa.util import (
    compare_with_tolerance,
    contextualize_text,
    get_inner_html_from_xpath,
    remove_markup,
    sanitize_html,
)


@ddt.ddt
class UtilTest(unittest.TestCase):
    """Tests for util"""

    def setUp(self):
        super().setUp()
        self.system = mock_capa_system()

    def test_compare_with_tolerance(self):  # pylint: disable=too-many-statements
        """Test numeric comparison with relative and absolute tolerances."""
        # Test default tolerance '0.001%' (it is relative)
        result = compare_with_tolerance(100.0, 100.0)
        assert result
        result = compare_with_tolerance(100.001, 100.0)
        assert result
        result = compare_with_tolerance(101.0, 100.0)
        assert not result
        # Test absolute percentage tolerance
        result = compare_with_tolerance(109.9, 100.0, "10%", False)
        assert result
        result = compare_with_tolerance(110.1, 100.0, "10%", False)
        assert not result
        # Test relative percentage tolerance
        result = compare_with_tolerance(111.0, 100.0, "10%", True)
        assert result
        result = compare_with_tolerance(112.0, 100.0, "10%", True)
        assert not result
        # Test absolute tolerance (string)
        result = compare_with_tolerance(109.9, 100.0, "10.0", False)
        assert result
        result = compare_with_tolerance(110.1, 100.0, "10.0", False)
        assert not result
        # Test relative tolerance (string)
        result = compare_with_tolerance(111.0, 100.0, "0.1", True)
        assert result
        result = compare_with_tolerance(112.0, 100.0, "0.1", True)
        assert not result
        # Test absolute tolerance (float)
        result = compare_with_tolerance(109.9, 100.0, 10.0, False)
        assert result
        result = compare_with_tolerance(110.1, 100.0, 10.0, False)
        assert not result
        # Test relative tolerance (float)
        result = compare_with_tolerance(111.0, 100.0, 0.1, True)
        assert result
        result = compare_with_tolerance(112.0, 100.0, 0.1, True)
        assert not result
        # Infinite values #
        infinity = float("Inf")
        # Test relative tolerance (float)
        result = compare_with_tolerance(infinity, 100.0, 1.0, True)
        assert not result
        result = compare_with_tolerance(100.0, infinity, 1.0, True)
        assert not result
        result = compare_with_tolerance(infinity, infinity, 1.0, True)
        assert result
        # Test absolute tolerance (float)
        result = compare_with_tolerance(infinity, 100.0, 1.0, False)
        assert not result
        result = compare_with_tolerance(100.0, infinity, 1.0, False)
        assert not result
        result = compare_with_tolerance(infinity, infinity, 1.0, False)
        assert result
        # Test relative tolerance (string)
        result = compare_with_tolerance(infinity, 100.0, "1.0", True)
        assert not result
        result = compare_with_tolerance(100.0, infinity, "1.0", True)
        assert not result
        result = compare_with_tolerance(infinity, infinity, "1.0", True)
        assert result
        # Test absolute tolerance (string)
        result = compare_with_tolerance(infinity, 100.0, "1.0", False)
        assert not result
        result = compare_with_tolerance(100.0, infinity, "1.0", False)
        assert not result
        result = compare_with_tolerance(infinity, infinity, "1.0", False)
        assert result
        # Test absolute tolerance for smaller values
        result = compare_with_tolerance(100.01, 100.0, 0.01, False)
        assert result
        result = compare_with_tolerance(100.001, 100.0, 0.001, False)
        assert result
        result = compare_with_tolerance(100.01, 100.0, "0.01%", False)
        assert result
        result = compare_with_tolerance(100.002, 100.0, 0.001, False)
        assert not result
        result = compare_with_tolerance(0.4, 0.44, 0.01, False)
        assert not result
        result = compare_with_tolerance(100.01, 100.0, 0.010, False)
        assert result

        # Test complex_number instructor_complex
        result = compare_with_tolerance(0.4, complex(0.44, 0), 0.01, False)
        assert not result
        result = compare_with_tolerance(100.01, complex(100.0, 0), 0.010, False)
        assert result
        result = compare_with_tolerance(110.1, complex(100.0, 0), "10.0", False)
        assert not result
        result = compare_with_tolerance(111.0, complex(100.0, 0), "10%", True)
        assert result

    def test_sanitize_html(self):
        """
        Test for html sanitization with nh3.
        """
        allowed_tags = ["div", "p", "audio", "pre", "span"]
        for tag in allowed_tags:
            queue_msg = f"<{tag}>Test message</{tag}>"
            assert sanitize_html(queue_msg) == queue_msg

        not_allowed_tag = "script"
        queue_msg = f"<{not_allowed_tag}>Test message</{not_allowed_tag}>"
        expected = ""
        assert sanitize_html(queue_msg) == expected

    def test_get_inner_html_from_xpath(self):
        """
        Test for getting inner html as string from xpath node.
        """
        xpath_node = etree.XML('<hint style="smtng">aa<a href="#">bb</a>cc</hint>')
        assert get_inner_html_from_xpath(xpath_node) == 'aa<a href="#">bb</a>cc'

    def test_remove_markup(self):
        """
        Test for markup removal with nh3.
        """
        assert (
            remove_markup("The <mark>Truth</mark> is <em>Out There</em> & you need to <strong>find</strong> it")
            == "The Truth is Out There &amp; you need to find it"
        )

    @ddt.data("When the root level failš the whole hierarchy won’t work anymore.", "あなたあなたあなた")
    def test_contextualize_text(self, context_value):
        """Verify that variable substitution works as intended with non-ascii characters."""
        key = "answer0"
        text = "$answer0"
        context = {key: context_value}
        contextual_text = contextualize_text(text, context)
        assert context_value == contextual_text

    def test_contextualize_text_with_non_ascii_context(self):
        """Verify that variable substitution works as intended with non-ascii characters."""
        key = "あなた$a $b"
        text = "$" + key
        context = {"a": "あなたあなたあなた", "b": "あなたhi"}
        expected_text = "$あなたあなたあなたあなた あなたhi"
        contextual_text = contextualize_text(text, context)
        assert expected_text == contextual_text


class UseUnsafeCodejail(TestContextDecorator):
    """
    Tell codejail to run in unsafe mode for the scope of the decorator.
    Use this as a decorator on Django TestCase classes or methods.

    This is needed because codejail has significant OS-level setup requirements
    which we don't even attempt to fulfill for unit testing purposes. Running
    tests in unsafe mode (that is, running code executions in-process, with no
    sandboxing) is only safe because we control the contents of the unit tests.
    It's not a perfect replica of how safe mode operates but it's generally good
    enough for testing the integration and overall behavior.
    """

    def __init__(self):
        self.old_be_unsafe = None
        super().__init__()

    def enable(self):
        """Enable unsafe mode for codejail within the test scope."""
        self.old_be_unsafe = codejail.safe_exec.ALWAYS_BE_UNSAFE
        codejail.safe_exec.ALWAYS_BE_UNSAFE = True

    def disable(self):
        """Restore the previous codejail unsafe mode state."""
        codejail.safe_exec.ALWAYS_BE_UNSAFE = self.old_be_unsafe



"""
LMS Interface to external queueing system (xqueue)
"""

import hashlib
import json
import logging
from typing import TYPE_CHECKING, Dict, Optional

import requests
from requests.auth import HTTPBasicAuth

from xblocks_contrib.problem.capa.xqueue_submission import XQueueInterfaceSubmission

if TYPE_CHECKING:
    from xblocks_contrib.problem.capa_block import ProblemBlock

log = logging.getLogger(__name__)
DATEFORMAT = "%Y%m%d%H%M%S"

XQUEUE_METRIC_NAME = "edxapp.xqueue"

# Wait time for response from Xqueue.
XQUEUE_TIMEOUT = 35  # seconds
CONNECT_TIMEOUT = 3.05  # seconds
READ_TIMEOUT = 10  # seconds


def make_hashkey(seed):
    """
    Generate a string key by hashing
    """
    h = hashlib.md5()
    h.update(str(seed).encode("latin-1"))
    return h.hexdigest()


def make_xheader(lms_callback_url, lms_key, queue_name):
    """
    Generate header for delivery and reply of queue request.

    Xqueue header is a JSON-serialized dict::

        { 'lms_callback_url': url to which xqueue will return the request (string),
          'lms_key': secret key used by LMS to protect its state (string),
          'queue_name': designate a specific queue within xqueue server, e.g. 'MITx-6.00x' (string)
        }
    """
    return json.dumps({"lms_callback_url": lms_callback_url, "lms_key": lms_key, "queue_name": queue_name})


def parse_xreply(xreply):
    """
    Parse the reply from xqueue. Messages are JSON-serialized dict::

        { 'return_code': 0 (success), 1 (fail)
          'content': Message from xqueue (string)
        }
    """
    try:
        xreply = json.loads(xreply)
    except ValueError as err:
        log.error(err)
        return (1, "unexpected reply from server")

    return_code = xreply["return_code"]
    content = xreply["content"]

    return (return_code, content)


class XQueueInterface:
    """Initializes the XQueue interface."""

    def __init__(  # pylint: disable=too-many-positional-arguments
        self,
        url: str,
        django_auth: Dict[str, str],
        requests_auth: Optional[HTTPBasicAuth] = None,
        block: "ProblemBlock" = None,
        use_submission_service: bool = False,
    ):
        """
        Initializes the XQueue interface.

        Args:
            url (str): The URL of the XQueue service.
            django_auth (Dict[str, str]): Authentication credentials for Django.
            requests_auth (Optional[HTTPBasicAuth], optional): Authentication for HTTP requests. Defaults to None.
            block ('ProblemBlock', optional): Added as a parameter only to extract the course_id
                to check the course waffle flag `send_to_submission_course.enable`.
                This can be removed after the legacy xqueue is deprecated. Defaults to None.
            use_submission_service (bool): If True, use the edx-submissions service instead of XQueue.
        """
        self.url = url
        self.auth = django_auth
        self.session = requests.Session()
        self.session.auth = requests_auth
        self.block = block
        self.submission = XQueueInterfaceSubmission(self.block)
        self.use_submission_service = use_submission_service

    def send_to_queue(self, header, body, files_to_upload=None):
        """
        Submit a request to xqueue.

        header: JSON-serialized dict in the format described in 'xqueue_interface.make_xheader'

        body: Serialized data for the receipient behind the queueing service. The operation of
              xqueue is agnostic to the contents of 'body'

        files_to_upload: List of file objects to be uploaded to xqueue along with queue request

        Returns (error_code, msg) where error_code != 0 indicates an error
        """

        # log the send to xqueue
        header_info = json.loads(header)
        queue_name = header_info.get("queue_name", "")  # pylint: disable=unused-variable

        # Attempt to send to queue
        error, msg = self._send_to_queue(header, body, files_to_upload)

        # Log in, then try again
        if error and (msg == "login_required"):
            error, content = self._login()
            if error != 0:
                # when the login fails
                log.debug("Failed to login to queue: %s", content)
                return (error, content)
            if files_to_upload is not None:
                # Need to rewind file pointers
                for f in files_to_upload:
                    f.seek(0)
            error, msg = self._send_to_queue(header, body, files_to_upload)

        return error, msg

    def _login(self):
        payload = {"username": self.auth["username"], "password": self.auth["password"]}
        return self._http_post(self.url + "/xqueue/login/", payload)

    def _send_to_queue(self, header, body, files_to_upload):
        """Send the problem submission to XQueue, handling legacy fallback and edX submission logic."""

        payload = {"xqueue_header": header, "xqueue_body": body}
        files = {}
        if files_to_upload is not None:
            for f in files_to_upload:
                files.update({f.name: f})

        if self.block is None:
            # XQueueInterface: if self.block is None, falling back to legacy xqueue submission.
            log.error(
                "Unexpected None block: falling back to legacy xqueue submission. "
                "This may indicate a problem with the xqueue transition."
            )
            return self._http_post(self.url + "/xqueue/submit/", payload, files=files)

        header_info = json.loads(header)
        queue_key = header_info["lms_key"]

        if self.use_submission_service:
            submission = self.submission.send_to_submission(  # pylint: disable=unused-variable
                header, body, queue_key, files
            )
            return None, ""

        return self._http_post(self.url + "/xqueue/submit/", payload, files=files)

    def _http_post(self, url, data, files=None):
        """Send an HTTP POST request and handle connection errors, timeouts, and unexpected status codes."""

        try:
            response = self.session.post(url, data=data, files=files, timeout=(CONNECT_TIMEOUT, READ_TIMEOUT))
        except requests.exceptions.ConnectionError as err:
            log.error(err)
            return 1, "cannot connect to server"

        except requests.exceptions.ReadTimeout as err:
            log.error(err)
            return 1, "failed to read from the server"

        if response.status_code not in [200]:
            return 1, f"unexpected HTTP status code [{response.status_code}]"

        return parse_xreply(response.text)



"""
Utility functions for capa.
"""

import logging
import re
from cmath import isinf, isnan
from decimal import Decimal

import nh3
from calc import evaluator
from lxml import etree

from xblocks_contrib.problem.markup import HTML

# -----------------------------------------------------------------------------
#
# Utility functions used in CAPA responsetypes
DEFAULT_TOLERANCE = "0.001%"
log = logging.getLogger(__name__)


def compare_with_tolerance(student_complex, instructor_complex, tolerance=DEFAULT_TOLERANCE, relative_tolerance=False):
    """
    Compare student_complex to instructor_complex with maximum tolerance tolerance.

     - student_complex    :  student result (float complex number)
     - instructor_complex    :  instructor result (float complex number)
     - tolerance   :  float, or string (representing a float or a percentage)
     - relative_tolerance: bool, to explicitly use passed tolerance as relative

     Note: when a tolerance is a percentage (i.e. '10%'), it will compute that
     percentage of the instructor result and yield a number.

     If relative_tolerance is set to False, it will use that value and the
     instructor result to define the bounds of valid student result:
     instructor_complex = 10, tolerance = '10%' will give [9.0, 11.0].

     If relative_tolerance is set to True, it will use that value and both
     instructor result and student result to define the bounds of valid student
     result:
     instructor_complex = 10, student_complex = 20, tolerance = '10%' will give
     [8.0, 12.0].
     This is typically used internally to compare float, with a
     DEFAULT_TOLERANCE = '0.001%'.

     Default tolerance of 1e-3% is added to compare two floats for
     near-equality (to handle machine representation errors).
     Default tolerance is relative, as the acceptable difference between two
     floats depends on the magnitude of the floats.
     (http://randomascii.wordpress.com/2012/02/25/comparing-floating-point-numbers-2012-edition/)
     Examples::

        In [183]: 0.000016 - 1.6*10**-5
        Out[183]: -3.3881317890172014e-21
        In [212]: 1.9e24 - 1.9*10**24
        Out[212]: 268435456.0
    """
    if isinstance(tolerance, str):
        if tolerance == DEFAULT_TOLERANCE:
            relative_tolerance = True
        if tolerance.endswith("%"):
            tolerance = evaluator({}, {}, tolerance[:-1]) * 0.01
            if not relative_tolerance:
                tolerance = tolerance * abs(instructor_complex)
        else:
            tolerance = evaluator({}, {}, tolerance)

    if relative_tolerance:
        tolerance = tolerance * max(abs(student_complex), abs(instructor_complex))

    if isinf(student_complex) or isinf(instructor_complex):
        # If an input is infinite, we can end up with `abs(student_complex-instructor_complex)` and
        # `tolerance` both equal to infinity. Then, below we would have
        # `inf <= inf` which is a fail. Instead, compare directly.
        return student_complex == instructor_complex

    # because student_complex and instructor_complex are not necessarily
    # complex here, we enforce it here:
    student_complex = complex(student_complex)
    instructor_complex = complex(instructor_complex)

    # if both the instructor and student input are real,
    # compare them as Decimals to avoid rounding errors
    if not (instructor_complex.imag or student_complex.imag):
        # if either of these are not a number, short circuit and return False
        if isnan(instructor_complex.real) or isnan(student_complex.real):
            return False
        student_decimal = Decimal(str(student_complex.real))
        instructor_decimal = Decimal(str(instructor_complex.real))
        tolerance_decimal = Decimal(str(tolerance))
        return abs(student_decimal - instructor_decimal) <= tolerance_decimal

    # v1 and v2 are, in general, complex numbers:
    # there are some notes about backward compatibility issue: see responsetypes.get_staff_ans()).
    return abs(student_complex - instructor_complex) <= tolerance


def contextualize_text(text, context):  # private
    """
    Takes a string with variables. E.g. $a+$b.
    Does a substitution of those variables from the context
    """

    def convert_to_str(value):
        """The method tries to convert unicode/non-ascii values into string"""
        try:
            return str(value)
        except UnicodeEncodeError:
            return value.encode("utf8", errors="ignore")

    if not text:
        return text

    for key in sorted(context, key=len, reverse=True):
        # TODO (vshnayder): This whole replacement thing is a big hack
        # right now--context contains not just the vars defined in the
        # program, but also e.g. a reference to the numpy module.
        # Should be a separate dict of variables that should be
        # replaced.
        context_key = "$" + key
        if context_key in (text.decode("utf-8") if isinstance(text, bytes) else text):
            text = convert_to_str(text)
            context_value = convert_to_str(context[key])
            text = text.replace(context_key, context_value)

    return text


def convert_files_to_filenames(answers):
    """
    Check for File objects in the dict of submitted answers,
        convert File objects to their filename (string)
    """
    new_answers = {}
    for answer_id in answers.keys():
        answer = answers[answer_id]
        # Files are stored as a list, even if one file
        if is_list_of_files(answer):
            new_answers[answer_id] = [f.name for f in answer]
        else:
            new_answers[answer_id] = answers[answer_id]
    return new_answers


def is_list_of_files(files):
    """Return True if the input is a list of valid files."""
    return isinstance(files, list) and all(is_file(f) for f in files)


def is_file(file_to_test):
    """
    Duck typing to check if 'file_to_test' is a File object
    """
    return all(hasattr(file_to_test, method) for method in ["read", "name"])


def find_with_default(node, path, default):
    """
    Look for a child of node using , and return its text if found.
    Otherwise returns default.

    Arguments:
       node: lxml node
       path: xpath search expression
       default: value to return if nothing found

    Returns:
       node.find(path).text if the find succeeds, default otherwise.

    """
    v = node.find(path)
    if v is not None:
        return v.text

    return default


def sanitize_html(html_code):
    """
    Sanitize html_code for safe embed on LMS pages.

    Used to sanitize XQueue responses from Matlab.
    """
    attributes = nh3.ALLOWED_ATTRIBUTES.copy()
    attributes.update(
        {
            "*": {"class", "style", "id"},
            "audio": {"controls", "autobuffer", "autoplay", "src"},
            "img": {"src", "width", "height", "class"},
        }
    )
    output = nh3.clean(
        html_code, tags=nh3.ALLOWED_TAGS | {"div", "p", "audio", "pre", "img", "span"}, attributes=attributes
    )
    return output


def get_inner_html_from_xpath(xpath_node):
    """
    Returns inner html as string from xpath node.

    """
    # returns string from xpath node
    html = etree.tostring(xpath_node).strip().decode("utf-8")
    # strips outer tag from html string
    # xss-lint: disable=python-interpolate-html
    inner_html = re.sub(f"(?ms)<{xpath_node.tag}[^>]*>(.*)</{xpath_node.tag}>", "\\1", html)
    return inner_html.strip()


def remove_markup(html):
    """
    Return html with markup stripped and text HTML-escaped.

    >>> nh3.clean("<b>Rock & Roll</b>", tags=set())
    'Rock &amp; Roll'
    >>> nh3.clean("<b>Rock &amp; Roll</b>", tags=set())
    'Rock &amp; Roll'
    """
    return HTML(nh3.clean(html, tags=set()))


def get_course_id_from_capa_block(capa_block):
    """
    Extract a stringified course run key from a CAPA block (aka ProblemBlock).

    This is a bit of a hack. Its intended use is to allow us to pass the course id
    (if available) to `safe_exec`, enabling course-run-specific resource limits
    in the safe execution environment (codejail).

    Arguments:
        capa_block (ProblemBlock|None)

    Returns: str|None
        The stringified course run key of the block.
        If not available, fall back to None.
    """
    if not capa_block:
        return None
    try:
        return str(capa_block.scope_ids.usage_id.course_key)
    except (AttributeError, TypeError):
        # AttributeError:
        #     If the capa block lacks scope ids or has unexpected scope ids, we
        #     would rather fall back to `None` than let an AttributeError be raised
        #     here.
        # TypeError:
        #     Old Mongo usage keys lack a 'run' specifier, and may
        #     raise a type error when we try to serialize them into a course
        #     run key. This is tolerable because such course runs are deprecated.
        return None



"""
Exceptions related to safe exec.
"""


class CodejailServiceParseError(Exception):
    """
    An exception that is raised whenever we have issues with data parsing.
    """


class CodejailServiceStatusError(Exception):
    """
    An exception that is raised whenever Codejail service response status is different to 200.
    """


class CodejailServiceUnavailable(Exception):
    """
    An exception that is raised whenever Codejail service is unavailable.
    """
